(* ======================================================================== *)
(*               Proof Procedures for Certified RCF Judgments               *)
(*                      - Derived proof procedures -                        *)
(*                                                                          *)
(*  by G.O.Passmore, LFCS, Univ. Edinburgh and Clare Hall, Univ. Cambridge  *)
(* Contact: (e) grant.passmore@cl.cam.ac.uk   (w) www.cl.cam.ac.uk/~gp351/. *)
(* ======================================================================== *)

structure CertRCFp : CertRCFp =
struct

open CertRCFk; (* LCF-style kernel for univariate RCF *)

(* A complete univariate RCF decision procedure using the proof system
   defined in the CertRCF kernel. Given an implicitly existential RCF
   formula Phi(x), we work to decide its status, returning either

   (Fml Phi(x)) -|- True (w : RealAlg.real_alg)

    or

   (Fml Phi(x)) -|- False. *)

fun univ_prove (fm : Formula.formula) =
    let val t0 = ra (Fml fm)
	val t1 = rd t0
	val tk = ref t1
	val wf = ref false
    in 
	while ((thm_r (!tk) <> False)
	       andalso not(!wf))
	      do
	      (case thm_r (!tk) of
		   Disj (_, []) => tk := rf (!tk)
		 | _ => tk := re (!tk))
	handle (Witness r) =>
	       (tk := rt (!tk);
		wf := true);
	(!tk)		
  end;

(* Attempt to refute an existential RCF formula, returning true if
   we're successful, false otherwise. This is the simplest interface
   through which MetiTarski will access this certified RCF module.
   Note that this throws away the real algebraic witnesses we compute
   when an input formula is satisfiable. We'll want to later retrieve
   these for, e.g., an extended real algebraic version of the model-
   sharing framework. *)

fun univ_refute (fm : Formula.formula) =
    case thm_r (univ_prove fm) of
	True _ => false
      | False => (print "CertRCFp.univ_refute succeeded!\n"; true)
      | _ => raise Useful.Error 
		   ("Univ_refute failed to decide its input formula. " ^
		    "This shouldn't be possible."); 

end
